# -*- coding: utf-8 -*-

"""
jishaku.inspections test
~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2021 Devon (scarletcafe) R
:license: MIT, see LICENSE for more details.

"""

import collections  # for __iadd__ test
import typing

import discord
import pytest

from jishaku.repl.inspections import all_inspections
from tests.utils import sentinel


@pytest.mark.parametrize(
    "target",
    [
        4,
        discord.Client,  # cover type subclasses
        tuple,  # cover many-subclass truncation
        [False, 1, "2", 3.0],  # cover content types
        collections.Counter,  # cover inplace operators
        sentinel  # cover current-working-directory inspections
    ]
)
def test_object_inspection(target: typing.Any):
    for _, _ in all_inspections(target):
        pass



# -*- coding: utf-8 -*-

"""
jishaku converter test
~~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2021 Devon (scarletcafe) R
:license: MIT, see LICENSE for more details.

"""

import inspect
from io import BytesIO

import pytest

from jishaku.paginators import FilePaginator, WrappedPaginator


def test_file_paginator():

    base_text = inspect.cleandoc("""
    #!/usr/bin/env python
    # -*- coding: utf-8 -*-
    pass  # \u3088\u308d\u3057\u304f
    """)

    # test standard encoding
    pages = FilePaginator(BytesIO(base_text.encode("utf-8"))).pages

    assert len(pages) == 1
    assert pages[0] == f"```python\n{base_text}\n```"

    # test linespan
    pages = FilePaginator(BytesIO(base_text.encode("utf-8")), line_span=(2, 2)).pages

    assert len(pages) == 1
    assert pages[0] == "```python\n# -*- coding: utf-8 -*-\n```"

    # test reception to encoding hint
    base_text = inspect.cleandoc("""
    #!/usr/bin/env python
    # -*- coding: cp932 -*-
    pass  # \u3088\u308d\u3057\u304f
    """)

    pages = FilePaginator(BytesIO(base_text.encode("cp932"))).pages

    assert len(pages) == 1
    assert pages[0] == f"```python\n{base_text}\n```"

    # test without encoding hint
    with pytest.raises(UnicodeDecodeError):
        FilePaginator(BytesIO("\u3088\u308d\u3057\u304f".encode("cp932")))

    # test with wrong encoding hint
    with pytest.raises(UnicodeDecodeError):
        FilePaginator(BytesIO("-*- coding: utf-8 -*-\n\u3088\u308d\u3057\u304f".encode("cp932")))

    # test OOB
    with pytest.raises(ValueError):
        FilePaginator(BytesIO("one\ntwo\nthree\nfour".encode('utf-8')), line_span=(-1, 20))


def test_wrapped_paginator():
    paginator = WrappedPaginator(max_size=200)
    paginator.add_line("abcde " * 50)
    assert len(paginator.pages) == 2

    paginator = WrappedPaginator(max_size=200, include_wrapped=False)
    paginator.add_line("abcde " * 50)
    assert len(paginator.pages) == 2

# TODO: Write test for interactions-based paginator interface



# -*- coding: utf-8 -*-

"""
jishaku.features.shell
~~~~~~~~~~~~~~~~~~~~~~~~

The jishaku shell commands.

:copyright: (c) 2021 Devon (scarletcafe) R
:license: MIT, see LICENSE for more details.

"""

import contextlib
import pathlib
import re
import shutil
import sys
import tempfile
import typing

import discord
from discord import ui
from discord.ext import commands

from jishaku.codeblocks import Codeblock, codeblock_converter
from jishaku.exception_handling import ReplResponseReactor
from jishaku.features.baseclass import Feature
from jishaku.flags import Flags
from jishaku.paginators import PaginatorInterface, WrappedPaginator
from jishaku.shell import ShellReader
from jishaku.types import ContextA

SCAFFOLD_FOLDER = pathlib.Path(__file__).parent / 'scaffolds'


@contextlib.contextmanager
def scaffold(name: str, **kwargs: typing.Any):
    """
    A context manager that sets up a temporary directory with a scaffold formatted to kwargs.

    This allows it to create temporary projects for different shell tools according to the template.
    """

    source = SCAFFOLD_FOLDER / name

    if not (source.exists() and source.is_dir()):
        raise ValueError(f"{name} is not a valid shell scaffold")

    with tempfile.TemporaryDirectory() as temp:
        temp = pathlib.Path(temp)

        for item in source.glob("**/*"):
            if '__pycache__' in str(item):
                continue

            if item.is_file():
                with open(item, 'r', encoding='utf-8') as fp:
                    content = fp.read()

                target = temp / item.relative_to(source)
                target.parent.mkdir(parents=True, exist_ok=True)

                with open(target, 'w', encoding='utf-8') as fp:
                    fp.write(content.format(**kwargs))

        yield temp


class ShellFeature(Feature):
    """
    Feature containing the shell-related commands
    """

    class ShellStandardInputModal(ui.Modal, title="Type input"):
        """Modal that prompts users for text to provide to stdin"""

        stdin_content: ui.TextInput[ui.Modal] = ui.TextInput(label="Text", style=discord.TextStyle.short)

        def __init__(self, reader: ShellReader, *args: typing.Any, **kwargs: typing.Any):
            super().__init__(*args, timeout=300, **kwargs)
            self.reader = reader

        async def on_submit(self, interaction: discord.Interaction, /):
            value = self.stdin_content.value or ""

            if self.reader.stdin and self.reader.stdin.writable():
                if sys.platform == "win32":
                    self.reader.stdin.write(f"{value}\r\n".encode('utf-8'))
                else:
                    self.reader.stdin.write(f"{value}\r".encode('utf-8'))
                self.reader.stdin.flush()

                await interaction.response.send_message(
                    content="Sent into stdin",
                    ephemeral=True
                )
            else:
                await interaction.response.send_message(
                    content="Stdin is not writable",
                    ephemeral=True
                )

    @Feature.Command(parent="jsk", name="shell", aliases=["bash", "sh", "powershell", "ps1", "ps", "cmd", "terminal"])
    async def jsk_shell(self, ctx: ContextA, *, argument: codeblock_converter):  # type: ignore
        """
        Executes statements in the system shell.

        This uses the system shell as defined in $SHELL, or `/bin/bash` otherwise.
        Execution can be cancelled by closing the paginator.
        """

        if typing.TYPE_CHECKING:
            argument: Codeblock = argument  # type: ignore

        async with ReplResponseReactor(ctx.message):
            with self.submit(ctx):
                with ShellReader(argument.content, escape_ansi=not Flags.use_ansi(ctx)) as reader:
                    prefix = "```" + reader.highlight

                    paginator = WrappedPaginator(prefix=prefix, max_size=1975)
                    paginator.add_line(f"{reader.ps1} {argument.content}\n")

                    async def send_standard_input(interaction: discord.Interaction):
                        await interaction.response.send_modal(self.ShellStandardInputModal(reader))

                    stdin_button = ui.Button(label="\N{KEYBOARD} Send standard input")
                    stdin_button.callback = send_standard_input

                    interface = PaginatorInterface(ctx.bot, paginator, owner=ctx.author, additional_buttons=[stdin_button])
                    self.bot.loop.create_task(interface.send_to(ctx))

                    async for line in reader:
                        if interface.closed:
                            return
                        await interface.add_line(line)

                await interface.add_line(f"\n[status] Return code {reader.close_code}")

    @Feature.Command(parent="jsk", name="git")
    async def jsk_git(self, ctx: ContextA, *, argument: codeblock_converter):  # type: ignore
        """
        Shortcut for 'jsk sh git'. Invokes the system shell.
        """

        if typing.TYPE_CHECKING:
            argument: Codeblock = argument  # type: ignore

        return await ctx.invoke(self.jsk_shell, argument=Codeblock(argument.language, "git " + argument.content))  # type: ignore

    @Feature.Command(parent="jsk", name="pip")
    async def jsk_pip(self, ctx: commands.Context, *, argument: codeblock_converter):  # type: ignore
        """
        Shortcut for 'jsk sh pip'. Invokes the system shell.
        """

        if typing.TYPE_CHECKING:
            argument: Codeblock = argument  # type: ignore

        executable: str = "pip"
        location = pathlib.Path(sys.prefix)

        for test in (
            location / 'bin' / 'pip',
            location / 'bin' / 'pip3',
            location / 'Scripts' / 'pip.exe',
            location / 'Scripts' / 'pip3.exe',
        ):
            if test.exists() and test.is_file():
                executable = str(test)
                break

        return await ctx.invoke(self.jsk_shell, argument=Codeblock(argument.language, f"{executable} {argument.content}"))  # type: ignore

    if shutil.which('node') and shutil.which('npm'):
        @Feature.Command(parent="jsk", name="node")
        async def jsk_node(self, ctx: commands.Context, *, argument: codeblock_converter):  # type: ignore
            """
            Shortcut for scaffolding and executing 'npm run'. Only exists if the executables are detected.
            """

            if typing.TYPE_CHECKING:
                argument: Codeblock = argument  # type: ignore

            requirements = ''.join(f"npm install {match} && " for match in re.findall('// jsk require: (.+)', argument.content))

            with scaffold('npm', content=argument.content) as directory:
                return await ctx.invoke(self.jsk_shell, argument=Codeblock("js", f"cd {directory} && {requirements}npm run main"))  # type: ignore

    if shutil.which('pyright'):
        @Feature.Command(parent="jsk", name="pyright")
        async def jsk_pyright(self, ctx: commands.Context, *, argument: codeblock_converter):  # type: ignore
            """
            Shortcut for scaffolding and executing 'pyright main.py'. Only exists if the executables are detected.
            """

            if typing.TYPE_CHECKING:
                argument: Codeblock = argument  # type: ignore

            with scaffold('pyright', content=argument.content) as directory:
                return await ctx.invoke(self.jsk_shell, argument=Codeblock("js", f"cd {directory} && pyright main.py"))  # type: ignore

    if shutil.which('rustc') and shutil.which('cargo'):
        @Feature.Command(parent="jsk", name="rustc")
        async def jsk_rustc(self, ctx: commands.Context, *, argument: codeblock_converter):  # type: ignore
            """
            Shortcut for scaffolding and executing 'cargo run'. Only exists if the executables are detected.
            """

            if typing.TYPE_CHECKING:
                argument: Codeblock = argument  # type: ignore

            requirements = '\n'.join(re.findall('// jsk require: (.+)', argument.content))

            with scaffold('cargo', content=argument.content, requirements=requirements) as directory:
                return await ctx.invoke(self.jsk_shell, argument=Codeblock("rust", f"cd {directory} && cargo run"))  # type: ignore



# -*- coding: utf-8 -*-

"""
jishaku
~~~~~~~

A discord.py extension including useful tools for bot development and debugging.

:copyright: (c) 2024 Devon (scarletcafe) R
:license: MIT, see LICENSE for more details.

"""

from jishaku.cog import Jishaku, STANDARD_FEATURES, OPTIONAL_FEATURES, setup
from jishaku.features.baseclass import Feature
from jishaku.flags import Flags
from jishaku.meta import *  # noqa: F403

__all__ = (
    'Jishaku',
    'Feature',
    'Flags',
    'STANDARD_FEATURES',
    'OPTIONAL_FEATURES',
    'setup',
)



# -*- coding: utf-8 -*-

"""
jishaku ast tree generation test
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2022 Devon (scarletcafe) R
:license: MIT, see LICENSE for more details.

"""

import inspect

from jishaku.repl.disassembly import create_tree


def test_ast_missing_fields():
    # should not raise
    create_tree(inspect.cleandoc("""
        def h(*, a):
            print(a)
    """), use_ansi=False)



# -*- coding: utf-8 -*-

"""
jishaku.formatting
~~~~~~~~~~~~~~~~~~~

Advanced formatting constructs

:copyright: (c) 2021 Devon (scarletcafe) R
:license: MIT, see LICENSE for more details.

"""

import dataclasses
import typing

# List of block characters:
#  - Dashed character from the left
#  - Left hand of span arm
#  - Middle bridge of span arm
#  - Right hand of span arm
#  - Extended part of arm on earlier lines to the right
#  - Extended part of arm on earlier lines to the left
BLOCK_CHARACTERS_SIMPLE = (
    '-', '|', '_', '|', '|', '|'
)

BLOCK_CHARACTERS_FANCY = (
    '-',
    '\N{BOX DRAWINGS LIGHT UP AND RIGHT}',
    '\N{BOX DRAWINGS LIGHT HORIZONTAL}',
    '\N{BOX DRAWINGS LIGHT UP AND LEFT}',
    '\N{BOX DRAWINGS LIGHT VERTICAL}',
    '\N{BOX DRAWINGS LIGHT VERTICAL}'
)


@dataclasses.dataclass
class LineAnnotation:
    """
    A line annotation contains:
    - Annotation text
    - A span (start and end point)
    - An (optional) ANSI code for the annotation text and line
    - An (optional) ANSI code for the annotated text's foreground
    - An (optional) ANSI code for the annotated text's background
    """

    text: str
    span: typing.Optional[typing.Tuple[int, int]]
    annotation_ansi: typing.Union[typing.Sequence[int], int, None] = None
    text_foreground: typing.Optional[int] = None
    text_background: typing.Optional[int] = None


class LineFormatter:
    """
    Class that formats a single line with a list of ordered annotations like so:

            one (two three) four
    FOO ‐‐‐‐└─┘ ││ │ │   ││ │  │
    BAR ‐‐‐‐‐‐‐‐└─────────┘ │  │
    BAZ ‐‐‐‐‐‐‐‐‐└─┘ │   │  │  │
    FOO ‐‐‐‐‐‐‐‐‐‐‐‐‐└───┘  │  │
    BAR ‐‐‐‐‐‐‐‐‐‐‐‐‐‐‐‐‐‐‐‐└──┘

    """
    def __init__(self, line: str):
        self.line: str = line
        self.max_annotation_size: int = 0
        self.annotations: typing.List[LineAnnotation] = []

    def add_annotation(
        self,
        text: str, span: typing.Optional[typing.Tuple[int, int]],
        annotation_ansi: typing.Union[typing.Sequence[int], int, None] = None,
        text_foreground: typing.Optional[int] = None,
        text_background: typing.Optional[int] = None
    ):
        """
        Add an annotation to the line formatter.
        The order in which the annotations are added matters.
        """

        # Flip the span order if right is earlier than left
        if span and span[0] > span[1]:
            span = (span[1], span[0])

        self.max_annotation_size = max(self.max_annotation_size, len(text))
        self.annotations.append(LineAnnotation(text, span, annotation_ansi, text_foreground, text_background))

    def output(self, use_complex: bool = True, use_ansi: bool = False) -> str:
        """
        Output the line and associated annotations, optionally with color
        """

        lines: typing.List[str] = []
        block = BLOCK_CHARACTERS_FANCY if use_complex else BLOCK_CHARACTERS_SIMPLE

        # The main line
        if use_ansi:  # pylint: disable=too-many-nested-blocks
            spans = [
                (annotation.span, annotation.text_foreground, annotation.text_background)
                for annotation in self.annotations
                if annotation.span and (annotation.text_foreground or annotation.text_background)
            ]

            if spans:
                # Sort spans by shortest and earliest first
                spans.sort(key=lambda s: (s[0][1] - s[0][0], s[0][0]))

                line = [" " * (self.max_annotation_size + 2)]
                color = (None, None)

                for index, character in enumerate(self.line):
                    foreground = None
                    background = None

                    for span in spans:
                        if span[0][0] <= index <= span[0][1]:
                            if foreground is None and span[1] is not None:
                                foreground = span[1]
                            if background is None and span[2] is not None:
                                background = span[2]

                        if foreground is not None and background is not None:
                            break

                    if (foreground, background) != color:
                        line.append(
                            "\u001b[0"
                            + (f';{foreground}' if foreground else '')
                            + (f';{background}' if background else '')
                            + 'm'
                            + character
                        )
                    else:
                        line.append(character)

                    color = (foreground, background)

                if color != (None, None):
                    line.append("\u001b[0m")

                lines.append("".join(line))

            else:
                # No spans, so optimize by just skipping bothering
                lines.append(" " * (self.max_annotation_size + 2) + self.line)
        else:
            lines.append(" " * (self.max_annotation_size + 2) + self.line)

        # Format the annotations
        annotation_strokes: typing.List[typing.Tuple[int, int, typing.Union[typing.Sequence[int], int, None]]] = []

        # Stroke generation pass
        for index, annotation in enumerate(self.annotations):
            if not annotation.text or not annotation.span:
                continue
            annotation_strokes.append((index, annotation.span[0], annotation.annotation_ansi))
            annotation_strokes.append((index, annotation.span[1], annotation.annotation_ansi))

        # Sort by stroke position, and then annotations so that earlier annotations take priority
        annotation_strokes.sort(key=lambda s: (s[1], s[0]))

        def to_ansi_text(ansi: typing.Union[typing.Sequence[int], int, None]) -> str:
            if ansi is None:
                return "\u001b[0m"
            if isinstance(ansi, int):
                return f"\u001b[0;{ansi}m"
            return f"\u001b[0;{';'.join(str(x) for x in ansi)}m"

        # Now generating the actual lines
        for index, annotation in enumerate(self.annotations):
            if not annotation.text:
                continue

            if use_ansi:
                line = [f"{to_ansi_text(annotation.annotation_ansi)}{annotation.text.rjust(self.max_annotation_size)} {block[0] if annotation.span else ' '}"]
            else:
                line = [f"{annotation.text.rjust(self.max_annotation_size)} {block[0] if annotation.span else ' '}"]

            # Go through each stroke exhaustively
            character_index = 0
            color = annotation.annotation_ansi

            for stroke in annotation_strokes:
                # Skip this if it's an earlier annotation or before the character index
                if stroke[0] < index or stroke[1] < character_index:
                    continue

                if use_ansi and annotation.span and character_index < annotation.span[1] and color != annotation.annotation_ansi:
                    line.append(to_ansi_text(annotation.annotation_ansi))
                    color = annotation.annotation_ansi

                line.append(
                    (
                        " " if not annotation.span else
                        block[0] if character_index < annotation.span[0] else
                        block[2] if character_index < annotation.span[1] else
                        " "
                    )
                    * (stroke[1] - character_index)
                )

                stroke_type = (
                    block[4] if not annotation.span else
                    block[5] if stroke[1] < annotation.span[0] else
                    block[3] if stroke[1] == annotation.span[1] else
                    block[1] if stroke[1] == annotation.span[0] else
                    block[2] if stroke[1] < annotation.span[1] else
                    block[4]
                )

                if use_ansi:
                    target_color = annotation.annotation_ansi if annotation.span and annotation.span[0] < stroke[1] < annotation.span[1] else stroke[2]
                    if target_color != color:
                        line.append(f"{to_ansi_text(target_color)}{stroke_type}")
                        color = target_color
                    else:
                        line.append(stroke_type)
                else:
                    line.append(stroke_type)

                character_index = stroke[1] + 1

            lines.append(''.join(line))

        return "\n".join(lines) + ("\u001b[0m" if use_ansi else "")


class MultilineFormatter:
    """
    A wrapper around LineFormatter that allows annotating a full block of text.
    """
    def __init__(self, text: str):
        self.lines: typing.List[LineFormatter] = [
            LineFormatter(line) for line in text.splitlines()
        ]

    def add_annotation(
        self,
        line: int,
        text: str, span: typing.Optional[typing.Tuple[int, int]],
        annotation_ansi: typing.Union[typing.Sequence[int], int, None] = None,
        text_foreground: typing.Optional[int] = None,
        text_background: typing.Optional[int] = None
    ):
        """
        Add an annotation to a line of this formatter.
        The order in which the annotations are added matters.
        """

        self.lines[line].add_annotation(text, span, annotation_ansi, text_foreground, text_background)

    def output(self, use_complex: bool = True, use_ansi: bool = False) -> str:
        """
        Outputs all lines with their annotations
        """
        max_line = max(formatter.max_annotation_size for formatter in self.lines)

        lines: typing.List[str] = []
        for formatter in self.lines:
            formatter.max_annotation_size = max_line
            lines.append(formatter.output(use_complex, use_ansi))

        return "\n".join(lines)



# -*- coding: utf-8 -*-

"""
jishaku.meta
~~~~~~~~~~~~

Meta information about jishaku.

:copyright: (c) 2021 Devon (scarletcafe) R
:license: MIT, see LICENSE for more details.

"""

import typing

__all__ = (
    '__author__',
    '__copyright__',
    '__docformat__',
    '__license__',
    '__title__',
    '__version__',
    'version_info'
)


class VersionInfo(typing.NamedTuple):
    """Version info named tuple for Jishaku"""
    major: int
    minor: int
    micro: int
    releaselevel: str
    serial: int


version_info = VersionInfo(major=2, minor=6, micro=1, releaselevel='final', serial=0)

__author__ = 'scarletcafe'
__copyright__ = 'Copyright 2021 Devon (scarletcafe) R'
__docformat__ = 'restructuredtext en'
__license__ = 'MIT'
__title__ = 'jishaku'
__version__ = '.'.join(map(str, (version_info.major, version_info.minor, version_info.micro)))
